# pyrcmd.py: ver 1.1.2 (2018/05/03)
  # Copyright (C) T. Yoshiizumi, 2018 All rights reserved.
import os, sys, platform
import subprocess
import tempfile
import re
import pandas as pd
import glob
import gc
from rpy2.robjects import r
from platform import python_version
if int(python_version()[0]) < 3:
    import StringIO
else:
    import io
from collections import OrderedDict
from chardet.universaldetector import UniversalDetector
from textwrap import dedent

if __name__ == 'pyrcmd':
    # global variables
    _default_charset = 'utf-8'
    if platform.system() == 'Windows':
        _default_charset = 'cp932'
    _charset = _default_charset
    rpt_with_include = False
    rexec_count = 0
    sink_file = None
    log_file = None
    rdata_file = None
    load = None
    if int(python_version()[0]) < 3:
        reload(sys)
        sys.setdefaultencoding(_charset)
    html_template = dedent('''\
        <html>
        <head>
        <meta http-equiv="Content-type" content="text/html; charset=%s" />
        <title>%s</title>
        </head>  
        <body>
        %s
        </body>
        </html>
        ''')

    # sets global values (_charset, _default_charset)
    def set_charset(charset = None, default_charset = None):
        if charset is not None:
            global _charset
            _charset = charset
        if default_charset is not None:
            global _default_charset
            _default_charset = default_charset
        if int(python_version()[0]) < 3:
            reload(sys)
            sys.setdefaultencoding(_charset)

    # sets global values (sink_file, log_file, etc.)
    def set_logs():
        global rexec_count, sink_file, log_file, rdata_file, load
        rexec_count = rexec_count + 1
        log_file = "pyrcmd_log%02d.txt" % rexec_count
        temp1 = tempfile.NamedTemporaryFile(delete=False)
        temp2 = tempfile.NamedTemporaryFile(delete=False)
        temp1.close()
        temp2.close()
        sink_file = re.sub(r'\\', '/', temp1.name)
        rdata_file = re.sub(r'\\', '/', temp2.name)
        load = 'load("%s")' % rdata_file

    # scrapes off CrLf codes
    def scrape_crlf(ss):
        ss = re.sub(r"\r", "", ss)
        ss = re.sub(r"\n\n+", "\n", ss)
        ss = ss.rstrip('\n')
        return ss

    # gets the file's charset(encoding)
    def get_charset(fname):
        detector = UniversalDetector()
        f = open(fname, mode='rb')
        for binary in f:
            detector.feed(binary)
            if detector.done:
                break
        f.close()
        detector.close()
        enc = detector.result['encoding']
        if enc is None:
            enc = _default_charset
        return enc

    # read text file (return unicode string)
    def read_text(filename, charset=''):
        try:
            flag = os.path.exists(filename)
        except:
            flag = False
        if not flag:
            return None
        if charset == '':
            charset = get_charset(filename)
        if int(python_version()[0]) < 3:
            f = open(filename, mode='r')
            data = f.read()
            f.close()
            data = unicode(data, charset)
        else:
            f = open(filename, mode='rt', encoding=charset)
            data = f.read()
            f.close()
        return data

    # write a text to a file
    def write_text(filename, text, charset=''):
        if charset == '':
            charset = _charset
        if int(python_version()[0]) < 3:
            text = text.encode(charset)
            with open(filename, "w") as f:
                f.write(text)
        else:
            with open(filename, "w", encoding=charset) as f:
                f.write(text)

    # get '*.R' in directory (environment 'RPT_WITH')
    def rpt_rfiles():
        res = ''
        rwd = os.getenv("RPT_WITH")
        if rwd is not None:
            wldc = rwd + "\\*.R"
            flist = glob.glob(wldc)
            for ff in flist:
                ff_cset = get_charset(ff)
                ff = re.sub(r"\\", "/", ff)
                res = res + 'source("%s", encoding="%s")' % (ff, ff_cset) + "\n"
        return res

    # remove variables on R (free the memory used)
    def r_clear():
        gc.collect()
        r('rm(list = ls())')
        r('gc()')
        gc.collect()

    # execute R command(R CMD BATCH)
    def rexec(rsf, options=None, log=True, sink=True, rdata=True):
        if options is None:
            options = '--no-save -q --internet2'
        set_logs()
        remove_log()
        cset = _charset
        if os.path.exists(rsf):
            cset = get_charset(rsf)
            rsf = read_text(rsf, cset)
        wstr = "options(encoding='%s')\n" % (cset)
        if rpt_with_include:
            wstr = wstr + rpt_rfiles()
        if sink:
            wstr = wstr + "sink(file='%s', split=T)\n" % (sink_file)
        rsf = wstr + rsf
        if rdata:
            wstr = 'save(list = ls(), file="%s")' % (rdata_file)
            rsf = rsf + "\n" + wstr

        temp = tempfile.NamedTemporaryFile(delete=False)
        temp.write(rsf.encode(cset))
        temp.close()
        cmd = "R CMD BATCH %s --encoding=%s %s %s" \
          % (options, cset, temp.name, log_file)
        proc = subprocess.call( cmd , shell=True)
        os.unlink(temp.name)
        if os.path.exists(rdata_file):
            r(load)
        sink_data = None
        if os.path.exists(sink_file):
            sink_data = read_text(sink_file)
            if sink_data is not None:
                sink_data = re.sub(r"\A\s+", "", sink_data)
                sink_data = re.sub(r"\n> proc\.time\(\)\n.+\Z", "",
                  sink_data, flags=(re.MULTILINE | re.DOTALL))
                sink_data = sink_data.rstrip('\n')
        remove_log('rdata', 'sink')
        if log == False:
            remove_log('log')
        return sink_data

    # remove files (log, sink, rdata)
    def remove_log(*args):
        if len(args) == 0:
            flist = [log_file, sink_file, rdata_file]
        else:
            dct = {'log': log_file, 'sink': sink_file, 'rdata': rdata_file}
            flist = list()
            for x in args:
                if x in dct:
                    flist.append(dct[x])
        for fname in flist:
            if os.path.exists(fname):
                os.remove(fname)

    # remove all files (log, sink, rdata)
    def remove_all():
        wldc = "./pyrcmd_*[0-9][0-9].*"
        flist = glob.glob(wldc)
        for fname in flist:
            if os.path.exists(fname):
                os.remove(fname)

    # open StringIO and return the object(sio)
    def sio_open(ss = ''):
        sio_scr2 = dedent('''\
            sio = StringIO.StringIO(ss)
            sio.seek(0)
            ''')
        sio_scr3 = dedent('''\
            sio = io.BytesIO(ss)
            sio.seek(0)
            ''')
        if int(python_version()[0]) < 3:
            exec(sio_scr2)
        else:
            dct = {'ss': ss.encode()}
            exec(sio_scr3, globals(), dct)
            sio = dct['sio']
        return sio

    # rpy2 object into string
    def rp2str(ss = ''):
        scr2 = dedent('''\
            sio = StringIO.StringIO()
            print >>sio, ss
            sio.seek(0)
            ss2 = sio.read()
            sio.close()
            ss2 = scrape_crlf(ss2)
            ''')
        scr3 = dedent('''\
            sio = io.StringIO()
            print(ss, file=sio)
            sio.seek(0)
            ss2 = sio.read()
            sio.close()
            ss2 = scrape_crlf(ss2)
            ''')
        if int(python_version()[0]) < 3:
            exec(scr2)
        else:
            dct = {'ss': ss}
            exec(scr3, globals(), dct)
            ss2 = dct['ss2']
        return ss2

    # rpy2 vector into list
    def vt2list(vect):
        vtname = vect.__class__.__name__
        vlist = None
        if vtname == 'FactorVector':
            l1 = vt2list(vect.levels)
            l2 = list(vect)
            vlist = list()
            for x in l2:
                vlist.append(l1[x-1])
        elif vtname in ['IntVector', 'FloatVector', 'BoolVector', 'StrVector']:
            try:
                vlist = list(vect)
            except:
                vlist = None
            if vlist is None:
                vlist = list()
                for i in range(1, len(vect)+1):
                    vs = rp2str(vect.rx2(i))
                    vs = re.sub(r'^\[\d+\] *', '', vs)
                    vlist.append(eval(vs))
        if vlist is not None:
            na_count = 0
            for i in range(0, len(vlist)):
                if vlist[i] == '':
                    na_count = na_count + 1
                    if na_count > 1:
                        vlist[i] = 'NA%d' % (na_count)
                    else:
                        vlist[i] = 'NA'
        return vlist

    # python list into pandas.DataFrame
    def list2pd(lobj):
        if lobj.__class__.__name__ == 'range':
            lobj = list(lobj)
        if lobj.__class__.__name__ != 'list':
            return None
        if lobj[0].__class__.__name__ == 'list':
            sep_s = '\t'
            ss = ''
            if int(python_version()[0]) < 3:
                for l in lobj:
                    ss = ss + sep_s.join(map(str, l)) + "\n"
            else:
                for l in lobj:
                    ss = ss + sep_s.join(list(map(str, l))) + "\n"
            sio_f = sio_open(ss)
            cset = 'utf-8'
            if int(python_version()[0]) < 3:
                cset = _default_charset
            if lobj[0][0] == '':
                pdobj = pd.read_csv(sio_f, encoding=cset,
                  index_col=0, sep=sep_s)
            else:
                pdobj = pd.read_csv(sio_f, encoding=cset, sep=sep_s)
            sio_f.close()
        else:
            ss = ''
            for l in lobj:
                ss = ss + str(l) + "\n"
            sio_f = sio_open(ss)
            cset = 'utf-8'
            if int(python_version()[0]) < 3:
                cset = _default_charset
            pdobj = pd.read_csv(sio_f, header=None, \
              names=['V1'], encoding=cset)
            sio_f.close()
        return pdobj

    # if possible, converts pandas.DataFrame to nums, characters or Series
    def simplify(w):
        if w.__class__.__name__ != 'DataFrame':
            return w
        lcol = list(w.columns)
        lrow = list(w.index)
        if len(lcol) == 1 and len(lrow) == 1:
            val = w.values[0][0]
            if lcol[0] == 'V1':
                sobj = val
            else:
                sobj = pd.Series({lcol[0]: val})
        elif len(lcol) == 1:
            vlist = list()
            for v in w.values:
                vlist.append(v[0])
            if lcol[0] == 'V1':
                sobj = pd.Series(vlist)
            else:
                sobj = pd.Series({lcol[0]: vlist})
        elif len(lrow) == 1:
            vlist = w.values[0]
            if lcol[0] == 'V1' and lcol[1] == 'V2':
                sobj = pd.Series(vlist)
            else:
                sobj = pd.Series(OrderedDict(zip(lcol, vlist)))
        else:
            sobj = w
        return sobj

    # rpy2 object into pandas object
    def rp2pd(robj):
        nrobj = robj.__class__.__name__
        drobj = dir(robj)
        names_str = 'names' in drobj and \
          robj.names.__class__.__name__ == 'StrVector'
        if 'to_csvfile' in drobj:
            row_vt = vt2list(robj.rownames)
            row_flag = False
            for i in range(0, len(row_vt)):
                if not row_vt[i] == str(i+1):
                    row_flag = True
                    break
            temp = tempfile.NamedTemporaryFile(delete=False)
            temp.close()
            robj.to_csvfile(temp.name, row_names=row_flag)
            ss = read_text(temp.name, _default_charset)
            os.unlink(temp.name)
            ss = scrape_crlf(ss)
            sio_f = sio_open(ss)
            pdobj = pd.read_csv(sio_f, encoding='utf-8')
            sio_f.close()
        elif nrobj == 'Matrix':
            mlist = list()
            rnames = vt2list(robj.rownames)
            if rnames is None:
                rnames = list()
            cnames = vt2list(robj.colnames)
            if cnames is None:
                cnames = list()
                for i in range(1, robj.ncol+1):
                    cnames.append('V%d' % i)
            if len(rnames) > 0:
                cnames.insert(0, '')
            mlist.append(cnames)
            for i in range(1, robj.nrow+1):
                clist = list()
                if len(rnames) > 0:
                    clist.append(rnames[i-1])
                for j in range(1, robj.ncol+1):
                    v = rp2str(robj.rx2(i,j))
                    v = re.sub(r'^\[\d+\] *', '', v)
                    clist.append(v)
                mlist.append(clist)
            pdobj = list2pd(mlist)
        elif nrobj == 'Array':
            dlist = vt2list(robj.dim)
            if len(dlist) == 1:
                n = dlist[0]
                try:
                    l1 = vt2list(robj.dimnames.rx2(1))
                except:
                    l1 = None
                if l1 is None:
                    l1 = list()
                    for i in range(1, n+1):
                        l1.append('V%d' % (i))
                l2 = list()
                for i in range(1, n+1):
                    vs = rp2str(robj.rx2(i))
                    vs = re.sub(r'^\[\d+\] *', '', vs)
                    l2.append(vs)
                pdobj = list2pd([l1, l2])
            else:
                pdobj = None
        elif nrobj == 'ListVector':
            ln = len(robj)
            nlist = vt2list(robj.names)
            if nlist is None:
                nlist = range(1, ln+1)
            else:
                for i in range(0, ln):
                    if re.search(r'^NA', nlist[i]):
                        nlist[i] = i+1
            ndtf = list2pd(nlist)
            nlist = list()
            for v in ndtf.values:
                nlist.append(v[0])
            pdlist = list()
            for n in range(1, len(nlist)+1):
                w = robj.rx2(n)
                pdlist.append(rp2pd(w))
            pdobj = OrderedDict(zip(nlist, pdlist))
        elif names_str:
            l1 = vt2list(robj.names)
            l2 = vt2list(robj)
            pdobj = list2pd([l1, l2])
        else:
            lrobj = vt2list(robj)
            if lrobj is not None:
                pdobj = list2pd(lrobj)
            else:
                pdobj = None
        return simplify(pdobj)

    # creates a markdown table by pandas.DataFrame
    def md_table(dtf, align=None, index=True):
        if align is None:
            align_l = list()
            for i in range(0, len(dtf.columns)):
                dtype = str(dtf[dtf.columns[i]].dtype)
                if re.search(r'^int|^unint|^float', dtype):
                    align_l.append('----:')
                else:
                    align_l.append('-----')
            if index:
                align_l.insert(0, '-----')
            align_s = "|" + "|".join(align_l) + "|"
        else:
            align_cl = list(align)
            align_l = list()
            clen = len(dtf.columns)
            if index:
                clen = clen + 1
            if len(align_cl) < clen:
                for i in range(len(align_cl), clen):
                    align_cl.append('!')
            for i in range(0, clen):
                align_c = align_cl[i]
                if align_c == 'l':
                    align_l.append(':----')
                elif align_c == 'c':
                    align_l.append('--:--')
                elif align_c == 'r':
                    align_l.append('----:')
                else:
                    align_l.append('-----')
            align_s = "|" + "|".join(align_l) + "|"
        csv_s = dtf.to_csv(sep="\t", index=index, encoding='utf-8')
        csv_s = csv_s.replace("\r", "\n")
        csv_s = csv_s.rstrip('\n')
        csv_s = re.sub(r'\|', '\\|', csv_s)
        csv_s = re.sub(r"\t", '|', csv_s)
        lines = csv_s.split("\n")
        md_l = list()
        for line in lines:
            line = re.sub(r"^", '|', line)
            line = re.sub(r"$", '|', line)
            md_l.append(line)
        md_l.insert(1, align_s)
        md_s = "\n".join(md_l)
        if int(python_version()[0]) < 3:
            md_s = unicode(md_s, 'utf-8')
        return md_s

    # remake OrderedDict for template formatting
    def format_od(rod, **kwargs):
        if not rod.__class__.__name__ in ['OrderedDict', 'dict']:
            return rod
        table_type = kwargs.pop('table_type', 'markdown')
        prefix = kwargs.pop('prefix', '')
        fod = OrderedDict()
        key_stock = dict()
        for key in rod.keys():
            key2 = re.sub(r'[^A-Za-z0-9_]', '_', '%s' % key)
            key2 = re.sub(r'_+$', '', key2)
            if re.search(r'^[0-9]', key2):
                key2 = '_' + key2
            if key2 in key_stock:
                key2 = '_' + key2
            dobj = rod[key]
            nobj = dobj.__class__.__name__
            if nobj == 'OrderedDict':
                fod[prefix + key2] = format_od(dobj, **kwargs)
                key_stock[key2] = True
            elif nobj == 'DataFrame':
                if table_type in ['markdown', 'md']:
                    fod[prefix + key2] = md_table(dobj, **kwargs)
                elif table_type in ['html', 'htm']:
                    fod[prefix + key2] = dobj.to_html(**kwargs)
                else:
                    fod[prefix + key2] = dobj
                key_stock[key2] = True
            elif nobj == 'Series':
                windex = dobj.index.values
                if windex[0] in [0, 1]:
                    fod[prefix + key2] = list()
                    key_stock[key2] = True
                    for idx in windex:
                        fod[prefix + key2].append(dobj[idx])
                else:
                    for idx in windex:
                        key3 = re.sub(r'[^A-Za-z0-9_]', '_', idx)
                        key3 = re.sub(r'_+$', '', key3)
                        if re.search(r'^[0-9]', key3):
                            key3 = '_' + key3
                        if key3 in key_stock:
                            key3 = key2 + '_' + key3
                        fod[prefix + key3] = dobj[idx]
                        key_stock[key3] = True
            else:
                fod[prefix + key2] = dobj
                key_stock[key2] = True
        return fod

    # print each element of the Dict object
    def print_dict(dct, prefix=''):
        if not dct.__class__.__name__ in ['dict', 'OrderedDict']:
            return
        for key in dct.keys():
            key2 = '%s%s' % (prefix, key)
            print(key2)
            print('(%s)' % dct[key].__class__.__name__)
            if not dct[key].__class__.__name__ in ['dict', 'OrderedDict']:
                print(dct[key])
            else:
                print_dict(dct[key], key2 + '.')
            print('')

